use crate::lib::entity::{Ant, Egg, Entities, Food, Queen};
use crate::lib::point::{astar, Point};
use crate::lib::screen::{BoardCommand, Screen};
use std::collections::HashSet;

#[derive(Clone)]
pub struct World {
    pub cleared: HashSet<Point>,
    pub food: HashSet<u32>,
}

impl World {
    pub fn new() -> World {
        World {
            cleared: HashSet::new(),
            food: HashSet::new(),
        }
    }

    pub fn clear(&mut self, pos: Point) {
        self.cleared.insert(pos);
    }

    // make sure that this is as simple & robust as possible
    // then, use this set of rules when digging
    // call astar from the suggested dig site to origin (or whatever "home" coordinate")
    // if route does not exist, do not dig
    pub fn is_valid_movement(&self, target: &Point, b: &Screen) -> bool {
        let safe = 
               !self.cleared.contains(&target.down())
            || !self.cleared.contains(&target.left())
            || !self.cleared.contains(&target.right())
            || !self.cleared.contains(&target.bldiag())
            || !self.cleared.contains(&target.uldiag())
            || !self.cleared.contains(&target.brdiag())
            || !self.cleared.contains(&target.urdiag());

        // can only go to target if
        // its above or left or right of a not cleared space

        /*
        if target.is_above(current_pos) {
            safe =
                (!self.cleared.contains(&target.left()) || !self.cleared.contains(&target.right()))
                || (!self.cleared.contains(&current_pos.right()) && self.cleared.contains(&target.right()))
                || (!self.cleared.contains(&current_pos.left()) && self.cleared.contains(&target.left()));
        }

        if target.is_adjacent(current_pos) {
            let target_down = target.down();
            if self.cleared.contains(&target_down) {
                safe = !self.cleared.contains(&target_down.left())
                    || !self.cleared.contains(&target_down.right());
            } else {
                safe = !self.cleared.contains(&target.down());
            }
        }

        if target.is_below(current_pos) {
            safe = true;
        }*/

        // of course, its only a valid move if you can walk there!
        b.is_in_bounds(target) && self.cleared.contains(target) && safe
    }

    pub fn is_cleared(&self, pos: &Point) -> bool {
        self.cleared.contains(pos)
    }

    pub fn is_safe_to_dig(&self, current_pos: &Point, target: &Point, b: &Screen) -> bool {
        // create a scenario in which this has been dug out
        let mut hypothetical_world = self.clone();
        hypothetical_world.clear(*target);
        // test if we can return from the target to the beginning
        // if yes, dig
        let result = astar(target, &Point(0, 0), Some(&hypothetical_world), Some(b));
        //let to_current_pos = astar(target, current_pos, Some(&hypothetical_world), Some(b));
        /*if target.is_above(current_pos) {
            safe = !self.cleared.contains(&target.up());
        }

        if target.is_adjacent(current_pos) {
            safe = !self.cleared.contains(&target.up()) && !self.cleared.contains(&target.down());
        }

        if target.is_below(current_pos) {
           safe = !self.cleared.contains(&target.down()) && !self.cleared.contains(&target.left()) && !self.cleared.contains(&target.right());
        }*/
        b.is_in_bounds(target) && result.is_ok()
    }

    pub fn get_diggable(&self, b: &Screen, pos: &Point) -> Vec<Point> {
        let moves = b.get_valid_movements(pos);
        moves
            .iter()
            .filter(|p| !self.is_cleared(p))
            .map(|p| p.clone())
            .collect()
    }

    pub fn get_valid_movements(
        &self,
        pos: &Point,
        b: &Screen,
        return_diggables: bool,
    ) -> Vec<Point> {
        let moves = b.get_valid_movements(pos);
        let mut ans: Vec<Point> = moves
            .iter()
            .filter(|p| self.is_valid_movement(p, b))
            .map(|p| p.clone())
            .collect();

        if return_diggables {
            ans.extend(self.get_diggable(b, pos));
        }

        ans
    }
}

pub fn render(e: &Entities, w: &World, b: &Screen) {
    for c in w.cleared.iter() {
        b.render(c, "x");
    }

    for a in e.data.values() {
        a.render(b);
    }
}

pub fn simulate(e: &mut Entities, w: &mut World, b: &mut Screen) {
    let cmds: Vec<BoardCommand> = e
        .data
        .values_mut()
        .map(|a| {
            a.plan(w);
            a.step(b, w)
        })
        .collect();

    for cmd in cmds {
        match cmd {
            BoardCommand::Dig(pos) => {
                w.clear(pos);
            }
            BoardCommand::LayEgg(pos, id) => {
                let egg = Egg::new(pos.0, pos.1, id);
                e.add_entity(&egg);
            }
            BoardCommand::Hatch(egg_id, queen_id) => {
                let egg = e.data.remove(&egg_id);
                let pos = egg.unwrap().get_position();

                let q = e.data.get_mut(&queen_id).unwrap();
                let queen: &mut Queen = q.downcast_mut::<Queen>().unwrap();
                queen.egg_count -= 1;
                let ant = Ant::new(pos.0, pos.1);
                e.add_entity(&ant);
            }
            BoardCommand::SpawnFood(pos) => {
                let food = Food::new(pos.0, pos.1);
                e.add_entity(&food);
            }
            BoardCommand::Noop => {}
        }
    }
}
